# INSERT LICENSE

import time
import navis
# from typing import Tuple, Optional, List, Union
from matplotlib import pyplot as plt
import numpy as np
import networkx as nx
from neuroboom.utils import check_valid_neuron_input, calc_cable
from itertools import chain
# import plotly as py
# from plotly.graph_objs import *
import plotly.graph_objs as go
from plotly.offline import plot, iplot


# from logging import
# This script contains functions for plotting dendrograms, static and interactive

# dendro

VALID_PROGS = ["fdp", "dot", "neato"]


# def plot_dendrogram(
#     x: Union[navis.TreeNeuron, navis.neuronlist.NeuronList],
#     heal_neuron: bool = True,
#     downsample_neuron: float = 0.0,
#     plot_connectors: bool = True,
#     connector_confidence: Tuple[float, float] = (0.0, 0.0),
#     highlight_connectors: Optional = None,
#     fragment: bool = False,
#     presyn_color: List[List[float]] = [[0.9, 0.0, 0.0]],
#     postsyn_color: List[List[float]] = [[0.0, 0.0, 0.9]],
#     highlight_connector_color: List[List[float]] = [[0.0, 0.9, 0.0]],
#     highlight_connector_size: int = 20,
#     presyn_size: float = 0.1,
#     postsyn_size: float = 0.1,
#     prog: str = "dot",
# ) -> None:
#
#     # Typing & Input sanity checking
#     x = check_valid_neuron_input(x)
#
#     assert isinstance(
#         connector_confidence, tuple
#     ), f"Need to pass a tuple for confidence values. You have passed: {connector_confidence}"
#     assert (
#         len(connector_confidence) == 2
#     ), """Need to pass a tuple containing two values for confidence. \n
#            The first value is the confidence threshold for presynapses. \n
#           The second value is the confidence threshold for postsynapses
#        """
#
#     assert (
#         prog in VALID_PROGS
#     ), f"Unknown program parameter!. Please specify one of: {VALID_PROGS}"
# =======

# This script contains functions for plotting dendrograms, static and interactive
# Create Graph Structure

def create_graph_structure(x, returned_object='graph', prog='dot'):
    """
    Takes a navis neuron and creates a graph layout

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    if not isinstance(x, (navis.TreeNeuron, navis.neuronlist.NeuronList)):
        raise ValueError('Need to pass a Navis Tree Neuron type')
    elif isinstance(x, navis.neuronlist.NeuronList):
        if len(x) > 1:
            raise ValueError('Need to pass a SINGLE neuron')
        else:
            x = x[0]

    valid_objects = ['graph', 'positions', 'graph_and_positions']
    if returned_object not in valid_objects:
        raise ValueError('Unknown object type to return!')

    valid_progs = ['fdp', 'dot', 'neato']
    if prog not in valid_progs:
        raise ValueError('Unknown program parameter!')

    print('Creating Graph Structure...')
    g = nx.DiGraph()
    g.add_nodes_from(x.nodes.node_id)
    for e in x.nodes[['node_id', 'parent_id', 'parent_dist']].values:
        # Skip root node
        if e[1] == -1:
            continue
        g.add_edge(int(e[0]), int(e[1]), len=e[2])

    if returned_object == 'graph':
        print('Returning graph only')
        return(g)

    elif returned_object == 'positions':
        print('Calculating layout...')
        pos = nx.nx_agraph.graphviz_layout(g, prog=prog)
        print('Returning positions only')
        return(pos)

    elif returned_object == 'graph_and_positions':
        print('Calculating layout...')
        pos = nx.nx_agraph.graphviz_layout(g, prog=prog)
        print('Returning graph and positions')
        return(g, pos)

# dendrogram


def plot_dendrogram(x, heal_neuron=True,
                    downsample_neuron=0.0,
                    plot_connectors=True,
                    connector_confidence=(0.0, 0.0),
                    highlight_connectors=None,
                    fragment=False,
                    presyn_color=[[.9, .0, .0]],
                    postsyn_color=[[.0, .0, .9]],
                    highlight_connector_color=[[.0, .9, .0]],
                    highlight_connector_size=20,
                    presyn_size=.1, postsyn_size=.1,
                    prog='dot'):

    """
    Takes a navis neuron and creates a graph layout

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    if not isinstance(x, (navis.TreeNeuron, navis.neuronlist.NeuronList)):
        raise ValueError('Need to pass a Navis Tree Neuron type')
    elif isinstance(x, navis.neuronlist.NeuronList):
        if len(x) > 1:
            raise ValueError('Need to pass a SINGLE Neuron')
        else:
            x = x[0]

    if not isinstance(connector_confidence, tuple):
        raise ValueError('Need to pass a tuple for confidence values')
    elif isinstance(connector_confidence, tuple):
        if len(connector_confidence) != 2:
            raise ValueError('''Need to pass a tuple containing two values for confidence.
                             \n The first value is the confidence threshold for presynapses.
                             \n The second value is the confidence threshold for postsynapses''')

    valid_progs = ['fdp', 'dot', 'neato']
    if prog not in valid_progs:
        raise ValueError('Unknown program parameter!')

    start = time.time()

    if heal_neuron:
        print("Healing Neuron...")
        navis.heal_fragmented_neuron(x, inplace=True)

    if any(connector_confidence) > 0.0:
        print(
            """Thresholding synapses: only considering
        presynapses above {} confidence and postsynapses above {}""".format(
                connector_confidence[0], connector_confidence[1]
            )
        )
        presyn_included = x.connectors[x.connectors.type == "pre"][
            x.connectors.confidence > connector_confidence[0]
        ].connector_id.tolist()
        postsyn_included = x.connectors[x.connectors.type == "post"][
            x.connectors.confidence > connector_confidence[1]
        ].connector_id.tolist()
        connectors_included = list(
            chain.from_iterable([presyn_included, postsyn_included])
        )
        x.connectors = x.connectors[x.connectors.connector_id.isin(connectors_included)]

    if downsample_neuron > 0:
        print("Downsampling neuron, factor = {}".format(downsample_neuron))
        x = navis.downsample_neuron(
            x,
            downsampling_factor=downsample_neuron,
            preserve_nodes=x.connectors.node_id.unique().tolist(),
        )

    if "parent_dist" not in x.nodes:
        print("Calculating cable length...")
        x = calc_cable(x, return_skdata=True)

    g, pos = create_graph_structure(x, returned_object="graph_and_positions", prog=prog)

    # Plotting tree with the above layout
    print("Plotting Tree...")
    nx.draw(g, pos, node_size=0, arrows=False, width=0.25)

    # Whether to add soma or not
    if not fragment:
        print("Plotting soma")
        plt.scatter([pos[x.soma][0]], [pos[x.soma][1]], s=40, c=[[0, 0, 0]], zorder=1)

    if plot_connectors is not False:
        print("Plotting connectors...")
        plt.scatter(
            [
                pos[tn][0]
                for tn in x.connectors[x.connectors.type == "pre"].node_id.values
            ],
            [
                pos[tn][1]
                for tn in x.connectors[x.connectors.type == "pre"].node_id.values
            ],
            c=presyn_color,
            zorder=2,
            s=presyn_size,
            linewidths=1,
        )

        plt.scatter(
            [
                pos[tn][0]
                for tn in x.connectors[x.connectors.type == "post"].node_id.values
            ],
            [
                pos[tn][1]
                for tn in x.connectors[x.connectors.type == "post"].node_id.values
            ],
            c=postsyn_color,
            zorder=2,
            s=postsyn_size,
            linewidths=1,
        )

    if highlight_connectors is not None:

        if isinstance(highlight_connectors, (list, np.ndarray)) is True:
            hl_cn_coords = np.array(
                [
                    pos[tn]
                    for tn in x.connectors[
                        x.connectors.connector_id.isin(highlight_connectors)
                    ].treenode_id
                ]
            )
            plt.scatter(
                hl_cn_coords[:, 0],
                hl_cn_coords[:, 1],
                s=highlight_connector_size,
                c=highlight_connector_color,
                zorder=3,
                linewidths=1,
            )

        elif isinstance(highlight_connectors, dict) is True:
            for cn in highlight_connectors:
                if cn in highlight_connectors:
                    if cn is None:
                        continue
                    if cn not in x.connectors.connector_id.values:
                        print(
                            "Connector {} is not present in the neuron / neuron fragment".format(
                                cn
                            )
                        )
                    hl_cn_coords = np.array(
                        [
                            pos[tn]
                            for tn in x.connectors[
                                x.connectors.connector_id == cn
                            ].treenode_id
                        ]
                    )
                    plt.scatter(
                        hl_cn_coords[:, 0],
                        hl_cn_coords[:, 1],
                        s=highlight_connector_size,
                        c=highlight_connectors[cn],
                        zorder=3,
                    )

        else:

            raise TypeError(
                "Unable to highlight connectors from data of type {}".format(
                    type(highlight_connectors)
                )
            )

    print("Completed in %is" % int(time.time() - start))


def interactive_dendrogram(
    z,
    heal_neuron=True,
    plot_connectors=True,
    highlight_connectors=None,
    in_volume=None,
    prog="dot",
    inscreen=True,
    filename=None,
):
    """
    Takes a navis neuron and creates a graph layout

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """
    z = check_valid_neuron_input(z)
    assert prog in VALID_PROGS, "Unknown program parameter!"

    if heal_neuron:
        z = navis.heal_fragmented_neuron(z)

    # save start time
    start = time.time()

    # Necessary for neato layouts for preservation of segment lengths

    if "parent_dist" not in z.nodes:
        z = calc_cable(z, return_skdata=True)

    # Generation of networkx diagram

    g, pos = create_graph_structure(z, returned_object="graph_and_positions", prog=prog)

    print("Now creating plotly graph...")

    # Convering networkx nodes for plotly
    # NODES

    x = []
    y = []
    node_info = []

    for n in g.nodes():
        x_, y_ = pos[n]

        x.append(x_)
        y.append(y_)
        node_info.append("Node ID: {}".format(n))

    node_trace = go.Scatter(
        x=x,
        y=y,
        mode="markers",
        text=node_info,
        hoverinfo="text",
        marker=go.scatter.Marker(showscale=False),
    )

    # EDGES

    xe = []
    ye = []

    for e in g.edges():

        x0, y0 = pos[e[0]]
        x1, y1 = pos[e[1]]

        xe += [x0, x1, None]
        ye += [y0, y1, None]

    edge_trace = go.Scatter(
        x=xe,
        y=ye,
        line=go.scatter.Line(width=1.0, color="#000"),
        hoverinfo="none",
        mode="lines",
    )

    # SOMA

    xs = []
    ys = []

    for n in g.nodes():
        if n != z.soma:
            continue
        elif n == z.soma:

            x__, y__ = pos[n]
            xs.append(x__)
            ys.append(y__)

        else:
            break

    soma_trace = go.Scatter(
        x=xs,
        y=ys,
        mode="markers",
        hoverinfo="text",
        marker=dict(size=20, color="rgb(0,0,0)"),
        text="Soma, treenode_id:{}".format(z.soma),
    )

    # CONNECTORS:
    # RELATION  = 0 ARE PRESYNAPSES, RELATION = 1 ARE POSTSYNAPSES

    if plot_connectors is False:
        presynapse_connector_trace = go.Scatter()
        postsynapse_connector_trace = go.Scatter()

    elif plot_connectors is True:
        presynapse_connector_list = list(
            z.connectors[z.connectors.type == "pre"].node_id.values
        )

        x_pre = []
        y_pre = []
        presynapse_connector_info = []

        for node in g.nodes():
            for tn in presynapse_connector_list:

                if node == tn:

                    x, y = pos[node]

                    x_pre.append(x)
                    y_pre.append(y)
                    presynapse_connector_info.append(
                        "Presynapse, connector_id: {}".format(tn)
                    )

        presynapse_connector_trace = go.Scatter(
            x=x_pre,
            y=y_pre,
            text=presynapse_connector_info,
            mode="markers",
            hoverinfo="text",
            marker=dict(size=10, color="rgb(0,255,0)"),
        )

        postsynapse_connectors_list = list(
            z.connectors[z.connectors.type == "post"].node_id.values
        )

        x_post = []
        y_post = []
        postsynapse_connector_info = []

        for node in g.nodes():
            for tn in postsynapse_connectors_list:

                if node == tn:

                    x, y = pos[node]

                    x_post.append(x)
                    y_post.append(y)

                    postsynapse_connector_info.append(
                        "Postsynapse, connector id: {}".format(tn)
                    )

        postsynapse_connector_trace = go.Scatter(
            x=x_post,
            y=y_post,
            text=postsynapse_connector_info,
            mode="markers",
            hoverinfo="text",
            marker=dict(size=10, color="rgb(0,0,255)"),
        )

    if highlight_connectors is None:

        HC_trace = go.Scatter()

    elif highlight_connectors is not None:

        HC_nodes = []

        for i in highlight_connectors:

            HC_nodes.append(
                z.connectors[z.connectors.connector_id == i].node_id.values[0]
            )

            HC_x = []
            HC_y = []

            HC_info = []

            for node in g.nodes():

                for tn in HC_nodes:

                    if node is tn:

                        x, y = pos[node]

                        HC_x.append(x)
                        HC_y.append(y)

                        HC_info.append(
                            "Connector of Interest, connector_id: {}, treenode_id: {}".format(
                                z.connectors[
                                    z.connectors.node_id == node
                                ].connector_id.values[0],
                                node,
                            )
                        )

            HC_trace = go.Scatter(
                x=HC_x,
                y=HC_y,
                text=HC_info,
                mode="markers",
                hoverinfo="text",
                marker=dict(size=15, color="rgb(238,0,255)"),
            )

    # Highlight the nodes that are in a particular volume

    if in_volume is not None:

        # Volume of interest

        res = navis.in_volume(z.nodes, volume=in_volume, mode="IN")

        z.nodes["IN_VOLUME"] = res

        x_VOI = []
        y_VOI = []

        VOI_info = []

        for nodes in g.nodes():

            for tn in list(z.nodes[z.nodes.IN_VOLUME == True].node_id.values):

                if nodes == tn:

                    x, y = pos[tn]

                    x_VOI.append(x)
                    y_VOI.append(y)

                    VOI_info.append("Treenode {} is in {} volume".format(tn, in_volume))

        in_volume_trace = go.Scatter(
            x=x_VOI,
            y=y_VOI,
            text=VOI_info,
            mode="markers",
            hoverinfo="text",
            marker=dict(size=5, color="rgb(35,119,0)"),
        )

    print("Creating Plotly Graph")

    fig = go.Figure(
        data=[
            edge_trace,
            node_trace,
            soma_trace,
            presynapse_connector_trace,
            postsynapse_connector_trace,
            HC_trace,
            in_volume_trace,
        ],
        layout=go.Layout(
            title="Plotly graph of {} with {} layout".format(z.name, prog),
            titlefont=dict(size=16),
            showlegend=False,
            hovermode="closest",
            margin=dict(b=20, l=50, r=5, t=40),
            annotations=[
                dict(showarrow=False, xref="paper", yref="paper", x=0.005, y=-0.002)
            ],
            xaxis=go.layout.XAxis(showgrid=False, zeroline=False, showticklabels=False),
            yaxis=go.layout.YAxis(showgrid=False, zeroline=False, showticklabels=False),
        ),
    )

    print("Time taken: {}".format(time.time() - start))

    if inscreen is True:

        return iplot(fig)

    else:

        return plot(fig, filename=filename)



# INSERT LICENSE


# This script contains utility functions

# Importing dependencies

import navis
import numpy as np
import networkx as nx
import pymaid

# Calculating cable length between nodes - this function does not exist for navis neurons


def calc_cable(x, return_skdata=False):

    """
    Calculates cable length of navis neurons

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------

    """

    if not isinstance(x, (navis.TreeNeuron, navis.neuronlist.NeuronList)):
        raise ValueError('Need to pass a Navis Tree Neuron type')
    elif isinstance(x, navis.neuronlist.NeuronList):
        if len(x) > 1:
            raise ValueError('Need to pass a SINGLE CatmaidNeuron')
        else:
            x = x[0]

    nodes = x.nodes[~x.nodes.parent_id.isnull()]
    tn_coords = nodes[['x', 'y', 'z']].values

    this_tn = nodes.set_index('node_id')
    parent_coords = this_tn.reindex(nodes.parent_id.values, ['x', 'y', 'z']).values

    w = np.sqrt(np.sum((tn_coords - parent_coords) ** 2, axis=1))

    if return_skdata:
        nodes.loc[~nodes.parent_id.isnull(), 'parent_dist'] = w / 1000
        x.nodes = nodes
        return(x)

    return np.sum(w[np.logical_not(np.isnan(w))]) / 1000


def topological_sorting_of_nodes(x, return_object=['list', 'dict']):

    """
    Topological sorting of treenodes. Nodes and their parents are adjacent in the treenode table

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    g = nx.DiGraph()
    g.add_node(x.nodes[x.nodes.type == 'root'].treenode_id.values[0])
    root = x.nodes[x.nodes.type == 'root'].treenode_id.values[0]
    nodes_to_add = [i for i in x.nodes.treenode_id.tolist() if i != root]
    g.add_nodes_from(nodes_to_add)

    for e in x.nodes[['treenode_id', 'parent_id']].values:

        if e[1] is None:

            continue

        else:

            g.add_edge(e[1], e[0])

    topological_sort = [i for i in nx.topological_sort(g)]

    if return_object == 'list':

        return(topological_sort)

    elif return_object == 'dict':

        topological_sort = dict(zip(topological_sort, range(0, len(topological_sort))))

        return(topological_sort)


######################################################################

def prepare_neuron(x, change_units=True, factor=1e3):
    """
    Prepares neuron for electrotonic modelling

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """
    node_sort = dict([(i, k) for i, k in zip(range(len(x.nodes)), pymaid.node_label_sorting(x))])
    node_sort_rev = {i: j for j, i in node_sort.items()}
    # x = pymaid.downsample_neuron(x, resampling_factor = float('inf'))
    # x = pymaid.guess_radius(x)
    x.nodes['rank'] = x.nodes.treenode_id.map(node_sort_rev).tolist()
    x.nodes.sort_values(by=['rank'], ascending=True, inplace=True)
    x.nodes.reset_index(drop=True, inplace=True)

    x = pymaid.calc_cable(x, return_skdata=True)

    if not change_units:
        return(x)
    else:
        x.nodes['x'] = x.nodes['x'] / factor
        x.nodes['y'] = x.nodes['y'] / factor
        x.nodes['z'] = x.nodes['z'] / factor
        x.nodes['radius'] = x.nodes['radius'] / factor
        x.nodes['parent_dist'] = x.nodes['parent_dist'] / factor
        return(x)


def pymaid_to_navis(x):
    """
    Converts pymaid neuron types to navis


    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    if isinstance(x, pymaid.core.CatmaidNeuron):
        pass
    else:
        return('Need to pass a Catmaid Neuron!')

    if isinstance(x, pymaid.core.CatmaidNeuronList):
        if len(x) > 1:
            return('Need to pass a single Catmaid Neuron')
        else:
            x = x[0]

    # Getting topological sort of pymaid neuron
    x.nodes['rank'] = x.nodes.treenode_id.map(topological_sorting_of_nodes(x, return_object='dict'))
    x.nodes.sort_values(by=['rank'], ascending=True, inplace=True)

    # Getting the graph object of pymaid neuron and passing that to an empty navis TreeNeuron object
    x_graph = x.graph
    navis_neuron = navis.TreeNeuron(x_graph)

    # Populating the xyz columns of nodes
    for i, j in enumerate(x.nodes[['x', 'y', 'z']].values):

        navis_neuron.nodes.loc[i, 'x'] = j[0]
        navis_neuron.nodes.loc[i, 'y'] = j[1]
        navis_neuron.nodes.loc[i, 'z'] = j[2]

    # Adding type column
    navis_neuron.nodes['type'] = x.nodes.type.copy()

    # Adding connectors
    navis_neuron.connectors = x.connectors.copy()
    navis_neuron.connectors = ['pre' if i == 0 else 'post' for i in navis_neuron.connectors.type]

    # Adding soma
    navis_neuron.soma = x.soma

    # Adding name
    navis_neuron.name = x.neuron_name

    return(navis_neuron)



# INSERT LICENSE

# This script contains statistical functions
import navis
from collections import Counter
import pandas as pd
import numpy as np
from scipy import stats
from tqdm import tqdm
import copy
import random
import navis.interfaces.neuprint as nvneu


def presynapse_focality(x,
                        heal_fragmented_neuron=True,
                        confidence_threshold=(0.9, 0.9),
                        num_synapses_threshold=1):

    """
    Finds the connections that are downstream of 'x', where the presynpases of 'x' are focalised

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    if not isinstance(x, (navis.TreeNeuron, navis.neuronlist.NeuronList)):
        raise ValueError('Need to pass a Navis Tree Neuron type')
    elif isinstance(x, navis.neuronlist.NeuronList):
        if len(x) > 1:
            raise ValueError('Need to pass a SINGLE CatmaidNeuron')
        else:
            x = x[0]

    if heal_fragmented_neuron is True:

        x = navis.heal_fragmented_neuron(x)

    # Getting the connector table of synapses where x.id is the source
    synapse_connections = navis.interfaces.neuprint.fetch_synapse_connections(source_criteria=x.id)
    synapse_connections.astype(object)
    synapse_connections['node_id'] = object

    for i, j in tqdm(enumerate(synapse_connections[['x_pre', 'y_pre', 'z_pre']].values)):

        equals = np.equal(j, x.presynapses[['x', 'y', 'z']].values)
        all_equals = [np.all(k) for k in equals]
        node_id = x.presynapses[all_equals].node_id.tolist()
        # node_id = x.presynapses[[np.all(k) for k in np.equal(j, x.presynapses[['x', 'y', 'z']].values)]].node_id.tolist()
        synapse_connections.at[i, 'node_id'] = node_id

    truth_list = [True if len(np.unique(i)) > 1 else False for i in synapse_connections.node_id.values]
    if synapse_connections[truth_list].shape[0] == 0:

        synapse_connections.node_id = [np.unique(k)[0] for k in synapse_connections.node_id.tolist()]

    else:

        return('There are synapses associated with multiple nodes!!!!')

    
    synapse_connections = synapse_connections[synapse_connections.confidence_pre > confidence_threshold[0]][synapse_connections.confidence_post > confidence_threshold[1]].copy()
    count = Counter(synapse_connections.bodyId_post.tolist())
    count = {k: v for k, v in sorted(count.items(), key=lambda item: item[1], reverse=True)}
    truth_list = [True if count[i] > num_synapses_threshold else False for i in synapse_connections.bodyId_post]
    synapse_connections = synapse_connections[truth_list].copy()

    df = pd.DataFrame()
    df['partner_neuron'] = synapse_connections.bodyId_post.unique().tolist()
    df['gT'] = ""
    df['significance_val'] = ""
    df['p_val'] = ""
    df['num_syn'] = [count[i] for i in df.partner_neuron]

    return(synapse_connections, df)


def postsynapse_focality(
    x,
    heal_fragmented_neuron=True,
    split_neuron=False,
    confidence_threshold=(0.9, 0.9),
    num_synapses_threshold=1):

    if not isinstance(x, (navis.TreeNeuron, navis.neuronlist.NeuronList)):
        raise ValueError('Need to pass a Navis Tree Neuron type')
    elif isinstance(x, navis.neuronlist.NeuronList):
        if len(x) > 1:
            raise ValueError('Need to pass a SINGLE CatmaidNeuron')
        else:
            x = x[0]

    if heal_fragmented_neuron is True:

        x = navis.heal_fragmented_neuron(x)

    # Getting the connector table of synapses where x.id is the source
    synapse_connections = navis.interfaces.neuprint.fetch_synapse_connections(target_criteria=x.id)
    synapse_connections.astype(object)
    synapse_connections['node_id'] = object

    for i, j in tqdm(enumerate(synapse_connections[['x_post', 'y_post', 'z_post']].values)):

        node_id = x.postsynapses[[np.all(k) for k in np.equal(j, x.postsynapses[['x', 'y', 'z']].values)]].node_id.tolist()
        synapse_connections.at[i, 'node_id'] = node_id

    the_truth = [True if len(np.unique(i)) > 1 else False for i in synapse_connections.node_id.values]
    if synapse_connections[the_truth].shape[0] == 0:
    # if synapse_connections[[True if len(np.unique(i)) > 1 else False for i in synapse_connections.node_id.values]].shape[0] == 0:
        synapse_connections.node_id = [np.unique(k)[0] for k in synapse_connections.node_id.tolist()]

    else:
        return('There are synapses associated with multiple nodes!!!!')

    synapse_connections = synapse_connections[synapse_connections.confidence_pre > confidence_threshold[0]][synapse_connections.confidence_post > confidence_threshold[1]].copy()
    count = Counter(synapse_connections.bodyId_pre.tolist())
    count = {k:v for k, v in sorted(count.items(), key=lambda item: item[1], reverse=True)}
    truth_list = [True if count[i] > num_synapses_threshold else False for i in synapse_connections.bodyId_pre]
    synapse_connections = synapse_connections[truth_list].copy()

    df = pd.DataFrame()
    df['partner_neuron'] = synapse_connections.bodyId_pre.unique().tolist()
    df['gT'] = ""
    df['significance_val'] = ""
    df['p_val'] = ""
    df['num_syn'] = [count[i] for i in df.partner_neuron]

    return(synapse_connections, df)


def permut_test(x,
                measuring_node,
                synapse_connections,
                relation=['presyn', 'postsyn'],
                num_iter=10,
                df=None,
                count=None):

    geo_mat = navis.geodesic_matrix(x, tn_ids=measuring_node)
    geo_mat = geo_mat.T
    geo_mat.sort_values(by=[measuring_node], ascending=True, inplace=True)

    for k, j in tqdm(enumerate(df.partner_neuron)):

        if relation == 'presyn':

            total_distribution = geo_mat[~geo_mat.index.isin(synapse_connections[synapse_connections.bodyId_post == j].node_id.tolist())]
            specific_distribution = geo_mat[geo_mat.index.isin(synapse_connections[synapse_connections.bodyId_post == j].node_id.tolist())]

        elif relation == 'postsyn':

            total_distribution = geo_mat[~geo_mat.index.isin(synapse_connections[synapse_connections.bodyId_pre == j].node_id.tolist())]
            specific_distribution = geo_mat[geo_mat.index.isin(synapse_connections[synapse_connections.bodyId_pre == j].node_id.tolist())]

        total_mean = np.average(total_distribution.values)
        specific_mean = np.average(specific_distribution.values)

        gT = np.abs(total_mean - specific_mean)
        df.iloc[k, 1] = gT

        pV = list(total_distribution.values) + list(specific_distribution.values)
        pV = [i[0] for i in pV]

        pS = copy.copy(pV)
        pD = []

        for i in range(0, num_iter):

            random.shuffle(pS)
            pD.append(np.abs(np.average(pS[0:int(len(pS) / 2)]) - np.average(pS[int(len(pS) / 2):])))

        p_val = len(np.where(pD >= gT)[0]) / num_iter
        df.iloc[k, 2] = p_val

        ttest = stats.ttest_ind(total_distribution.values, specific_distribution.values)
        df.iloc[k, 3] = ttest[1][0]
        # if np.isnan(ttest[1]):
        #    df.iloc[k, 3] = np.nan
        # else:
        #    df.iloc[k, 3] = ttest[1][0]
    # df['num_syn'] = [count[i] for i in df.partner_neuron]
    df.sort_values(by=['p_val'], ascending=True, inplace=True)
    df.reset_index(inplace=True)
    return(df)

def permutation_test_complete(
                            x,
                            n_iter=10,
                            remove_fragments=True):

    a_pre, a_df = presynapse_focality(x, heal_fragmented_neuron=False, confidence_threshold=(0.9, 0.9))
    b_post, b_df = postsynapse_focality(x, heal_fragmented_neuron=False, confidence_threshold=(0.9, 0.9))

    presyn_pt = permut_test(x, x.root[0], synapse_connections=a_pre, relation='presyn', num_iter=n_iter, df=a_df)
    postsyn_pt = permut_test(x, x.root[0], synapse_connections=b_post, relation='postsyn', num_iter=n_iter, df=b_df)

    a_partner_neurons, a_roi = nvneu.fetch_neurons(presyn_pt.partner_neuron.tolist())
    b_partner_neurons, b_roi = nvneu.fetch_neurons(postsyn_pt.partner_neuron.tolist())

    a_type_dict = dict(zip(a_partner_neurons.bodyId.tolist(), a_partner_neurons.type.tolist()))
    b_type_dict = dict(zip(b_partner_neurons.bodyId.tolist(), b_partner_neurons.type.tolist()))

    partner_dict = {**a_type_dict, **b_type_dict}

    presyn_pt['partner_type'] = presyn_pt.partner_neuron.map(partner_dict, na_action='ignore')
    postsyn_pt['partner_type'] = postsyn_pt.partner_neuron.map(partner_dict, na_action='ignore')

    if remove_fragments:

        presyn_pt = presyn_pt[~presyn_pt.partner_type.isnull()].copy()
        postsyn_pt = postsyn_pt[~postsyn_pt.partner_type.isnull()].copy()

    return(presyn_pt, postsyn_pt)



# INSERT LICENSE

import scipy.sparse as sparse
import numpy as np

# Contains functions to model steady state passive electrotonic properties of neurons


def calculate_M_mat(x, Ra, Rm, Cm):
    """
    Calculates the voltage matrix

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """
    tn_to_index = dict(zip(x.nodes.treenode_id, x.nodes.index))
    nofcomps = x.nodes.shape[0]
    M = np.zeros((nofcomps, nofcomps))
    compdiam = x.nodes.radius * 2
    complength = np.zeros(nofcomps)

    # Skip root node
    for i in range(1, nofcomps):

        aind = int(x.nodes.treenode_id[i])
        bind = int(x.nodes.parent_id[i])

        axyz = x.nodes[x.nodes.treenode_id == aind][['x', 'y', 'z']].values
        bxyz = x.nodes[x.nodes.treenode_id == bind][['x', 'y', 'z']].values

        complength[i] = np.sqrt(np.sum((axyz - bxyz)**2))
        a_ind_nodes = x.nodes[x.nodes.treenode_id == aind].radius.values
        b_ind_nodes = x.nodes[x.nodes.treenode_id == bind].radius.values
        meandiam = (a_ind_nodes + b_ind_nodes) * .5
        # meandiam = (x.nodes[x.nodes.treenode_id == aind].radius.values + x.nodes[x.nodes.treenode_id == bind].radius.values) * .5
        area = (meandiam**2) / (4.0 * np.pi)

        M[tn_to_index[bind], tn_to_index[aind]] = -area / complength[i] / Ra*10**(-4)
        M[tn_to_index[aind], tn_to_index[bind]] = M[tn_to_index[bind], tn_to_index[aind]]

    complength[0] = complength[1]

    gleak = (compdiam * np.pi * complength) / (Rm * 10 ** 8)
    memcap = (compdiam * np.pi * complength) * Cm*(10 ** -6) / (10**8)

    for i in range(nofcomps):
        M[i, i] = gleak[i] - np.sum(M[i])

    M = sparse.csr_matrix(M)

    return(M, memcap)



# INSERT LICENSE
import navis.interfaces.neuprint as nvneu
from collections import Counter
import pandas as pd
import navis

# Create an adjacency matrix from synapse connections (neuprint)


def adjx_from_syn_conn(x, roi, rename_index=False):
    """
    Creates an adjacency matrixx from synapse connections

    Parameters
    ----------
    x : A navis neuron object

    returned_object : graph, graph_and_positions, positions

    prog : The layout type, can be dot, neato or fdp

    Returns
    -------
    graph : skeleton nodes as graph nodes with links between them as edges
    positions : a dictionary of the positions on the 2D plane of each node

    Examples
    --------
    """

    presyn_con = nvneu.fetch_synapse_connections(source_criteria=x)
    postsyn_con = nvneu.fetch_synapse_connections(target_criteria=x)

    pre_tt = navis.in_volume(presyn_con[['x_pre', 'y_pre', 'z_pre']].values, roi)
    post_tt = navis.in_volume(postsyn_con[['x_pre', 'y_pre', 'z_pre']].values, roi)

    roi_presyn_con = presyn_con[pre_tt].copy()
    roi_postsyn_con = postsyn_con[post_tt].copy()

    postsyn_roi_neurons = roi_presyn_con.bodyId_post.unique()

    n, _ = nvneu.fetch_neurons(postsyn_roi_neurons)
    partner_type_dict = dict(zip(n.bodyId.tolist(), n.type.tolist()))

    count = Counter(roi_presyn_con.bodyId_post)
    count = count.most_common()
    count_dict = dict(count)

    df = pd.DataFrame(columns=[x], index=[i for i, j in count], data=[count_dict[i] for i, j in count])

    if rename_index:

        df.index = [partner_type_dict[i] for i in df.index]

    return(df, partner_type_dict)
